;*************************************************************************
; Title	:    I2C (Single) Master Implementation
; Author:    Peter Fleury <pfleury@gmx.ch>  http://jump.to/fleury
;            based on Atmel Appl. Note AVR300
; File:      $Id: i2cmaster.S,v 1.11 2003/10/16 18:16:07 peter Exp $
; Software:  AVR-GCC 3.3
; Target:    any AVR device
;
; DESCRIPTION
; 	Basic routines for communicating with I2C slave devices. This
;	"single" master implementation is limited to one bus master on the
;	I2C bus. 
;  
;       Based on the Atmel Application Note AVR300, corrected and adapted 
;       to GNU assembler and AVR-GCC C call interface
;       Replaced the incorrect quarter period delays found in AVR300 with 
;       half period delays. 
;
; USAGE
;	These routines can be called from C, refere to file i2cmaster.h.
;       See example test_i2cmaster.c 
; 	Adapt the SCL and SDA port and pin definitions and eventually 
;	the delay routine to your target !
; 	Use 4.7k pull-up resistor on the SDA and SCL pin.
;
; NOTES
;	The I2C routines can be called either from non-interrupt or
;	interrupt routines, not both.
;
;*************************************************************************

#if (__GNUC__ * 100 + __GNUC_MINOR__) < 303
#error "This library requires AVR-GCC 3.3 or later, update to newer AVR-GCC compiler !"
#endif


#include <avr/io.h>



;***** Adapt these SCA and SCL port and pin definition to your target !!
;
#define SDA    	5		// SDA Port E, Pin 5   
#define SCL		4		// SCL Port E, Pin 4
#define SDA_PORT        PORTE           // SDA Port E
#define SCL_PORT        PORTE           // SCL Port E         

;******

;-- map the IO register back into the IO address space
#define SDA_DDR		(_SFR_IO_ADDR(SDA_PORT) - 1)
#define SCL_DDR		(_SFR_IO_ADDR(SCL_PORT) - 1)
#define SDA_OUT		_SFR_IO_ADDR(SDA_PORT)
#define SCL_OUT		_SFR_IO_ADDR(SCL_PORT)
#define SDA_IN		(_SFR_IO_ADDR(SDA_PORT) - 2)
#define SCL_IN		(_SFR_IO_ADDR(SCL_PORT) - 2)


#ifndef __tmp_reg__
#define __tmp_reg__ 0
#endif


	.section .text

;*************************************************************************
; delay half period
; For I2C in normal mode (100kHz), use T/2 > 5us
; For I2C in fast mode (400kHz),   use T/2 > 1.3us
;*************************************************************************
	.stabs	"",100,0,0,si2c_delay_T2
	.stabs	"SI2cMaster.S",100,0,0,si2c_delay_T2
	.func si2c_delay_T2	; delay 5.0 microsec with 8 Mhz crystal	
si2c_delay_T2:        ; 4 cycles
	rjmp 1f      ; 2   "
1:	rjmp 2f      ; 2   "
2:	rjmp 3f      ; 2   "
3:	rjmp 4f      ; 2   "
4:	rjmp 5f      ; 2   "
5:  rjmp 6f      ; 2   "
6:  rjmp 7f      ; 2   "
7:  rjmp 8f      ; 2   "
8:  rjmp 9f      ; 2   "
9:  rjmp 10f     ; 2   "
10: rjmp 11f     ; 2   "
11: rjmp 12f     ; 2   "
12: rjmp 16f     ; 2   "
13: rjmp 14f     ; 2   "
14: rjmp 15f     ; 2   "
15: rjmp 16f     ; 2   "

16: rjmp 17f      ; 2   "
17: rjmp 18f      ; 2   "
18: rjmp 19f      ; 2   "
19: rjmp 20f      ; 2   "
20: rjmp 21f      ; 2   "
21: rjmp 22f      ; 2   "
22: rjmp 23f      ; 2   "
23: rjmp 24f      ; 2   "
24: rjmp 25f     ; 2   "
25: rjmp 26f     ; 2   "
26: rjmp 27f     ; 2   "
27: rjmp 28f     ; 2   "
28: rjmp 29f     ; 2   "
29: rjmp 30f     ; 2   "
30: rjmp 31f     ; 2   "
31: rjmp 32f     ; 2   "


32: rjmp 33f     ; 2   "
33: rjmp 34f     ; 2   "
34: rjmp 35f     ; 2   "
35: rjmp 36f     ; 2   "
36: rjmp 37f     ; 2   "
37: rjmp 38f     ; 2   "
38: rjmp 39f     ; 2   "
39: rjmp 40f     ; 2   "
40: rjmp 41f     ; 2   "
41: rjmp 42f     ; 2   "
42: rjmp 43f     ; 2   "
43: rjmp 44f     ; 2   "
44: rjmp 45f     ; 2   "
45: rjmp 46f     ; 2   "
46: rjmp 47f     ; 2   "
47: rjmp 48f     ; 2   "
48: rjmp 49f     ; 2   "
49: rjmp 50f     ; 2   "
50: rjmp 51f     ; 2   "
51: rjmp 52f     ; 2   "
52: rjmp 53f     ; 2   "
53: rjmp 54f     ; 2   "
54: rjmp 55f     ; 2   "
55: rjmp 56f     ; 2   "

56: rjmp 57f     ; 2   "
57: rjmp 58f     ; 2   "
58: rjmp 59f     ; 2   "
59: rjmp 60f     ; 2   "
60: rjmp 61f     ; 2   "
61: rjmp 62f     ; 2   "
62: rjmp 63f     ; 2   "
63: rjmp 64f     ; 2   "
64: rjmp 65f     ; 2   "
65: rjmp 66f     ; 2   "
66: rjmp 67f     ; 2   "
67: rjmp 68f     ; 2   "
68: rjmp 69f     ; 2   "
69: rjmp 70f     ; 2   "
70: rjmp 71f     ; 2   "
71: rjmp 72f     ; 2   "
72: rjmp 73f     ; 2   "
73: rjmp 74f     ; 2   "
74: rjmp 75f     ; 2   "
75: rjmp 76f     ; 2   "

76: rjmp 77f     ; 2   "
77: rjmp 78f     ; 2   "
78: rjmp 79f     ; 2   "
79: rjmp 80f     ; 2   "
80: rjmp 81f     ; 2   "
81: rjmp 82f     ; 2   "
82: rjmp 83f     ; 2   "
83: rjmp 84f     ; 2   "
84: rjmp 85f     ; 2   "
85: rjmp 86f     ; 2   "
86: rjmp 87f     ; 2   "
87: rjmp 88f     ; 2   "
88: rjmp 89f     ; 2   "
89: rjmp 90f     ; 2   "
90: rjmp 91f     ; 2   "
91: rjmp 92f     ; 2   "
92: rjmp 93f     ; 2   "
93: rjmp 94f     ; 2   "
94: rjmp 95f     ; 2   "
95: rjmp 96f     ; 2   "

96: rjmp 97f     ; 2   "
97: rjmp 98f     ; 2   "
98: rjmp 99f     ; 2   "
99: rjmp 100f     ; 2   "
100: rjmp 101f     ; 2   "
101: rjmp 102f     ; 2   "
102: rjmp 103f     ; 2   "
103: rjmp 104f     ; 2   "
104: rjmp 105f     ; 2   "
105: rjmp 106f     ; 2   "
106: rjmp 107f     ; 2   "
107: rjmp 108f     ; 2   "
108: rjmp 109f     ; 2   "
109: rjmp 110f     ; 2   "
110: rjmp 111f     ; 2   "
111: rjmp 112f     ; 2   "
112: rjmp 113f     ; 2   "
113: rjmp 114f     ; 2   "
114: rjmp 115f     ; 2   "
115: rjmp 116f     ; 2   "

116: nop          ; 1   "
 ret          ; 3   "
	.endfunc     ; total 40 cyles = 5.0 microsec with 8 Mhz crystal 


;*************************************************************************
; Initialization of the I2C bus interface. Need to be called only once
; 
; extern void si2c_init(void)
;*************************************************************************
	.global si2c_init
	.func si2c_init
si2c_init:
	cbi SDA_DDR,SDA		;release SDA
	cbi SCL_DDR,SCL		;release SCL
	cbi SDA_OUT,SDA
	cbi SCL_OUT,SCL
	ret
	.endfunc


;*************************************************************************	
; Issues a start condition and sends address and transfer direction.
; return 0 = device accessible, 1= failed to access device
;
; extern unsigned char si2c_start(unsigned char addr);
;	addr = r24, return = r25(=0):r24
;*************************************************************************

	.global si2c_start
	.func   si2c_start
si2c_start:
	sbi 	SDA_DDR,SDA	;force SDA low
	rcall 	si2c_delay_T2	;delay T/2
	
	rcall 	si2c_write	;write address
	ret
	.endfunc		


;*************************************************************************
; Issues a repeated start condition and sends address and transfer direction.
; return 0 = device accessible, 1= failed to access device
;
; extern unsigned char si2c_rep_start(unsigned char addr);
;	addr = r24,  return = r25(=0):r24
;*************************************************************************

	.global si2c_rep_start
	.func	si2c_rep_start
si2c_rep_start:
	sbi	SCL_DDR,SCL	;force SCL low
	rcall 	si2c_delay_T2	;delay  T/2
	cbi	SDA_DDR,SDA	;release SDA
	rcall	si2c_delay_T2	;delay T/2
	cbi	SCL_DDR,SCL	;release SCL
	rcall 	si2c_delay_T2	;delay  T/2
	sbi 	SDA_DDR,SDA	;force SDA low
	rcall 	si2c_delay_T2	;delay	T/2
	
	rcall	si2c_write	;write address
	ret
	.endfunc


;*************************************************************************	
; Issues a start condition and sends address and transfer direction.
; If device is busy, use ack polling to wait until device is ready
;
; extern void si2c_start_wait(unsigned char addr);
;	addr = r24
;*************************************************************************

	.global si2c_start_wait
	.func   si2c_start_wait
si2c_start_wait:
	mov	__tmp_reg__,r24
si2c_start_wait1:
	sbi 	SDA_DDR,SDA	;force SDA low
	rcall 	si2c_delay_T2	;delay T/2
	mov	r24,__tmp_reg__
	rcall 	si2c_write	;write address
	tst	r24		;if device not busy -> done
	breq	si2c_start_wait_done
	rcall	si2c_stop	;terminate write operation
	rjmp	si2c_start_wait1	;device busy, poll ack again
si2c_start_wait_done:
	ret
	.endfunc	


;*************************************************************************
; Terminates the data transfer and releases the I2C bus
;
; extern void si2c_stop(void)
;*************************************************************************

	.global	si2c_stop
	.func	si2c_stop
si2c_stop:
	sbi	SCL_DDR,SCL	;force SCL low
	sbi	SDA_DDR,SDA	;force SDA low
	rcall	si2c_delay_T2	;delay T/2
	cbi	SCL_DDR,SCL	;release SCL
	rcall	si2c_delay_T2	;delay T/2
	cbi	SDA_DDR,SDA	;release SDA
	rcall	si2c_delay_T2	;delay T/2
	ret
	.endfunc


;*************************************************************************
; Send one byte to I2C device
; return 0 = write successful, 1 = write failed
;
; extern unsigned char si2c_write( unsigned char data );
;	data = r24,  return = r25(=0):r24
;*************************************************************************
	.global si2c_write
	.func	si2c_write
si2c_write:
	sec			;set carry flag
	rol 	r24		;shift in carry and out bit one
	rjmp	si2c_write_first
si2c_write_bit:
	lsl	r24		;if transmit register empty
si2c_write_first:
	breq	si2c_get_ack
	sbi	SCL_DDR,SCL	;force SCL low
	brcc	si2c_write_low
	nop
	cbi	SDA_DDR,SDA	;release SDA
	rjmp	si2c_write_high
si2c_write_low:
	sbi	SDA_DDR,SDA	;force SDA low
	rjmp	si2c_write_high
si2c_write_high:
	rcall 	si2c_delay_T2	;delay T/2
	cbi	SCL_DDR,SCL	;release SCL
	rcall	si2c_delay_T2	;delay T/2
	rjmp	si2c_write_bit
	
si2c_get_ack:
	sbi	SCL_DDR,SCL	;force SCL low
	cbi	SDA_DDR,SDA	;release SDA
	rcall	si2c_delay_T2	;delay T/2
	cbi	SCL_DDR,SCL	;release SCL
si2c_ack_wait:
	sbis	SCL_IN,SCL	;wait SCL high (in case wait states are inserted)
	rjmp	si2c_ack_wait
	
	clr	r24		;return 0
	sbic	SDA_IN,SDA	;if SDA high -> return 1
	ldi	r24,1
	rcall	si2c_delay_T2	;delay T/2
	clr	r25
	ret
	.endfunc



;*************************************************************************
; read one byte from the I2C device, send ack or nak to device
; (ack=1, send ack, request more data from device 
;  ack=0, send nak, read is followed by a stop condition)
;
; extern unsigned char si2c_read(unsigned char ack);
;	ack = r24, return = r25(=0):r24
; extern unsigned char si2c_readAck(void);
; extern unsigned char si2c_readNak(void);
; 	return = r25(=0):r24
;*************************************************************************
	.global si2c_readAck
	.global si2c_readNak
	.global si2c_read		
	.func	si2c_read
si2c_readNak:
	clr	r24
	rjmp	si2c_read
si2c_readAck:
	ldi	r24,0x01
si2c_read:
	ldi	r23,0x01	;data = 0x01
si2c_read_bit:
	sbi	SCL_DDR,SCL	;force SCL low
	cbi	SDA_DDR,SDA	;release SDA (from previous ACK)
	rcall	si2c_delay_T2	;delay T/2
	
	cbi	SCL_DDR,SCL	;release SCL
	rcall	si2c_delay_T2	;delay T/2
	
	clc			;clear carry flag
	sbic	SDA_IN,SDA	;if SDA is high
	sec			;  set carry flag
	
	rol	r23		;store bit
	brcc	si2c_read_bit	;while receive register not full
	
si2c_put_ack:
	sbi	SCL_DDR,SCL	;force SCL low	
	cpi	r24,1
	breq	si2c_put_ack_low	;if (ack=0)
	cbi	SDA_DDR,SDA	;      release SDA
	rjmp	si2c_put_ack_high
si2c_put_ack_low:                ;else
	sbi	SDA_DDR,SDA	;      force SDA low
si2c_put_ack_high:
	rcall	si2c_delay_T2	;delay T/2
	cbi	SCL_DDR,SCL	;release SCL
si2c_put_ack_wait:
	sbis	SCL_IN,SCL	;wait SCL high
	rjmp	si2c_put_ack_wait
	rcall	si2c_delay_T2	;delay T/2
	mov	r24,r23
	clr	r25
	ret
	.endfunc

